/* popmail (non-daemon Version) 4.4                            */
/* Copyright (c) 2001-2018 TODA, Takashi                       */
/*                                                             */
/* Original Version: popmail 1.2                               */
/* Copyright (c) 1995 Mitch Shimada(shimada@summit.novell.com) */
/*   TUXEDO System Group, Novell Japan.                        */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <sys/types.h>
#ifdef	linux
#include <unistd.h> /* necessary for getpass() */
#endif
#ifdef	WIN32
#include <io.h>
#include <winsock.h>
#include <conio.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#endif
#include <time.h>
#include <errno.h>
#ifdef	MSDOS
#include <dos.h>
#endif

#ifdef	WIN32
#define EWOULDBLOCK	WSAEWOULDBLOCK
#define	perror(routine)		fprintf(stderr, "%s: Error Code %d\n", routine, WSAGetLastError())
#endif

#ifdef	MSDOS
#define	O_NDELAY	FNDELAY
#endif

#ifdef	MSDOS
#define	DOSFILESYSTEM
#endif
#ifdef	WIN32
#define	DOSFILESYSTEM
#endif

/* Machine-Dependent Options */

/* Validate if you want to keep too many messages on the Server */
#define	KEEPMANY

/* Validate if you want to make result file in EUC-JP */
/* #define	EUC_JP */

/* Set to 1 if you want to wait for Server's reply
   before quitting after wrong password. Otherwise, set to 0 */
#define	WAITAFTERWRONGPSWD	0

/* Threshold of KEEPMANY */
#define	TOOMANY	100

/* Name of POP Protocol defined in /etc/services */
#define	PROTOCOL	"pop3"

/* Name of Result File */
#define	MBOX	"mbox"

/* Name of Buffering File */
#define	WBOX	"wbox"

/* Size of Socket Buffer */
#define SOCKBUFSIZE		4096

/* Size of Line Buffer */
#define	LINEBUFSIZE		32766
#ifdef	EUC_JP
#define	LINELIMIT	(LINEBUFSIZE / 2)
#else
#define	LINELIMIT	LINEBUFSIZE
#endif

/* Size of User-Type Buffer */
#ifdef	DOSFILESYSTEM
#define	TYPEBUFSIZE	16
#endif

/* End of Machine-Dependent Options */

/* File Name Delimiter */
#ifdef	DOSFILESYSTEM
#define	DELIMITER	"\\"
#else
#define	DELIMITER	"/"
#endif

/* Global Variables */
static	FILE	*fpwbox, *fpmbox;
static	char	*wboxname, *mboxname;
static	char	*message, *sockbuf;
static	char	*mboxopen = "w";
static	char	*hostname;
static	int	sock;

/* Command Line Options */
static	int	verbose = 0;
static	int	my_optind;	/* in stead of 'optind' with getopt() */

/* Prototype Definitions */
static	void	mimehead(char *);
static	void	kanjicnv(char *);
static	int	headcopy(char *, char *, int);
static	int	ConnectPopServer(char *);
static	void	ClosePopServer(int);
static	void	downloadEmail(int);
static	int	isEOM(void);
static	int	GetReply(void);
static	void	SendCommand(void);
static	char	*ReceiveCommand(void);
static	void	mboxpurge(void);
static	void	timestamp(char *);
static	void	setdecimal(char *, int, int, int);
#ifdef	DOSFILESYSTEM
static	char	*UserReply(const char *, int);
#endif

/*
 * Main procedure
 */
int
main(int argc, char *argv[]){
char	headbuf[81];
char	*loginname, *usr_reply, *allocptr;
int	msgnum, msgamount;
int	filetop, inheader, found_status;
int	subcont, fromcont, found_reply;
#ifdef	WIN32
WSADATA	wsaData;
#endif

#ifdef	WIN32
	if( WSAStartup(MAKEWORD(1, 1), &wsaData)){
		fprintf(stderr, "WinSock unavailiable\n");
		exit(126);
	}
#endif

	/*
	 * Allocating Transmit Buffer Memory
	 */
	if( (message = malloc(LINEBUFSIZE)) == NULL ){
		fprintf(stderr, "Cannot Allocate Transmit Line Buffer.\n");
		exit(126);
	}
	if( (sockbuf = malloc(SOCKBUFSIZE)) == NULL ){
		fprintf(stderr, "Cannot Allocate Transmit Socket Buffer.\n");
		exit(126);
	}

	/*
	 * Process the options
	 *
	 * The Origial Version is written by using getopt()
	 * but some of the author's targets do not support it. 
	 */
	my_optind = 1;
	if( argc <= 1 ){
		my_optind = -3;
	}
	else if( argv[1][0] == '-' ){
		if( argv[1][2] ){
			my_optind = -3;
		}
		else if( argv[1][1] == '-' ){
			my_optind++;
		}
		else if( argv[1][1] == 'v' ){
			verbose++;
			my_optind++;
		}
		else{
			my_optind = -3;
		}
	}
	if( (argc > my_optind + 2) || (argc <= my_optind) ){
		fprintf( stderr, "Usage: %s [ -v ] hostname [ loginname ]\n", argv[0]);
		exit(4);
	}

	/* display copyright... eliminated */

	/*
	 * Creating MailBox FileNames
	 */
	mboxname = getenv("HOME");
	wboxname = message;
	if( mboxname ){
		strcpy(wboxname, mboxname);
	}
	else{
		wboxname[0] = 0;
	}
	if( *wboxname ){
		if( strcmp(wboxname + strlen(wboxname) - 1, DELIMITER) ){
			strcat(wboxname, DELIMITER);
		}
	}
	mboxname = wboxname + strlen(wboxname) + strlen(WBOX) + 1;
	mboxname[0] = 0;
	strcpy(mboxname, wboxname);
	strcat(wboxname, WBOX);
	strcat(mboxname, MBOX);
	if( (allocptr = malloc(strlen(wboxname) + strlen(mboxname) + 2)) == NULL ){
		fprintf(stderr, "Cannot Allocate Memory for File Names.\n");
		exit(126);
	}
	memcpy(allocptr, message, strlen(wboxname) + strlen(mboxname) + 2);
	wboxname = allocptr;
	mboxname = wboxname + strlen(wboxname) + 1;

	/*
	 * Checking MailBox Files
	 */
	if(( fpwbox = fopen(wboxname, "r")) != NULL ){
		sprintf(message, "Buffering File '%s' Exists. Overwrite? (Y/N)", wboxname);
#ifdef	DOSFILESYSTEM
		usr_reply = UserReply(message, 1);
#else
		fputs(message, stderr);
		gets(message);
		usr_reply = message;
#endif
		if( (usr_reply[0] != 'Y') && (usr_reply[0] != 'y') ){
			exit(4);
		}
		fclose(fpwbox);
	}

	if( (fpmbox = fopen(mboxname, "r")) != NULL ){
		sprintf(message, "Mail Box File '%s' Exists. Overwrite or Append? (O/A/N)", mboxname);
#ifdef	DOSFILESYSTEM
		usr_reply = UserReply(message, 1);
#else
		fputs(message, stderr);
		gets(message);
		usr_reply = message;
#endif
		if( usr_reply[0] == 'A' || usr_reply[0] == 'a' ){
			mboxopen = "a";
		}
		else if( usr_reply[0] != 'O' && usr_reply[0] != 'o' ){
			exit(4);
		}
		fclose(fpmbox);
	}

	if( (fpwbox = fopen(wboxname, "w")) == NULL ){
		fprintf(stderr, "cannot open Buffering File '%s'.\n", wboxname);
		exit(126);
	}
	if( (fpmbox = fopen(mboxname, mboxopen)) == NULL ){
		fprintf(stderr, "cannot open Mail Box File '%s'.\n", mboxname);
		fclose(fpwbox);
		unlink(wboxname);
		exit(126);
	}

	/*
	 * Connect to the POP Server
	 */
	hostname = argv[my_optind];
	if( (sock = ConnectPopServer(hostname)) < 0 ){
		fprintf(stderr, "cannot connect Pop Server.\n");
		mboxpurge();
		exit(126);
	}
	if( GetReply() ){
		fputs("No Welcome Message : ", stderr);
		fputs(message, stderr);
		putc('\n', stderr);
		mboxpurge();
		ClosePopServer(0);
		exit(126);
	}

	/*
	 * Login Process
	 */
	/* You can allow to retry for wrong password 
	   only by making this process into infinite loop 
	   and changing the sub-process after sending password
	   into the iterating procedure (in case of failure) 
	   with the loop-escape (in case of success). 
	   However, if your server do not allow to retry, 
	   he may disconnect in excuting next 'send password' 
	   and this program can be aborted 
	   without executing 'mboxpurge();'. */

	if( argc > my_optind + 1 ){
		loginname = argv[my_optind + 1];
	}
	else{
#ifdef	DOSFILESYSTEM
		loginname = UserReply("login: ", 1);
#else
		fputs("login: ", stderr);
		gets(message);
		loginname = message + strlen(message) * 2 + 8;
		/* 8 == strlen("USER ") + some */
		strcpy(loginname, message);
#endif
	}
	if( *loginname == '\0' ){
		mboxpurge();
		ClosePopServer(1);
		exit(4);
	}
	strcpy(message, "USER ");
	strcat(message, loginname);
	SendCommand();
	if( GetReply() ){
		fputs("Cannot send login name : ", stderr);
		fputs(message, stderr);
		putc('\n', stderr);
		mboxpurge();
		ClosePopServer(1);
		exit(126);
	}
#ifdef	DOSFILESYSTEM
	usr_reply = UserReply("Password:", 0);
#else
	usr_reply = getpass("Password:");
#endif
	if( *usr_reply == '\0' ){
		mboxpurge();
		ClosePopServer(1);
		exit(4);
	}
	strcpy(message, "PASS ");
	strcat(message, usr_reply);
	SendCommand();
	if( GetReply() ){
		fputs(message, stderr);
		putc('\n', stderr);
		mboxpurge();
		ClosePopServer(WAITAFTERWRONGPSWD);
		exit(4);
	}

	/*
	 * How many does he have messages for me ?
	 */
	strcpy(message, "STAT");
	SendCommand();
	if( GetReply() ){
		fputs("cannot get status : ", stderr);
		fputs(message, stderr);
		putc('\n', stderr);
		mboxpurge();
		ClosePopServer(1);
		exit(126);
	}
	msgamount = atoi(strtok(message, " "));
	if( msgamount == 0 ){
		fprintf(stderr, "no mail on %s.\n", hostname);
		mboxpurge();
		ClosePopServer(1);
		return	0;
	}
	fprintf(stderr, "%d messages in %s.\n", msgamount, hostname);
#ifdef	KEEPMANY
	if( msgamount > TOOMANY ){
		msgamount = TOOMANY;
	}
#endif

	/*
	 * download messages
	 */
	for( msgnum = 0 ; msgnum < msgamount ; ++msgnum ){
		downloadEmail(msgnum + 1);
	}

	/*
	 * convert text-code in the messages and display their subjects
	 */
	fclose(fpwbox);
	if(( fpwbox = fopen(wboxname, "r")) == NULL ){
		fprintf(stderr, "cannot reopen Buffering File '%s'.\n", wboxname);
		exit(126);
	}
	msgnum = 0;
	filetop = 1;
	inheader = 0;
	found_status = 0;
	memset(headbuf, ' ', 79);
	headbuf[79] = '\n';
	headbuf[80] = 0;
	subcont = 0;
	fromcont = 0;
	found_reply = 0;
	while( fgets(message, LINEBUFSIZE, fpwbox) ){
		if( inheader && (*message == '\n') ){
			if( found_status == 0 ){
				fputs("Status: RO\n", fpmbox);
			}
			fputs(headbuf, stdout);
			memset(headbuf, ' ', 79);
			inheader = 0;
		}
		if( memcmp(message, "From ", 5) == 0 ){
			if( inheader ){
				if( found_status == 0 ){
					fputs("Status: RO\n", fpmbox);
				}
				fputs(headbuf, stdout);
				memset(headbuf, ' ', 79);
			}
			if( filetop == 0 ){
				msgnum++;
				/*fputs("\n", fpmbox) */
			}
			headbuf[0] = (msgnum / 10) + '0';
			headbuf[1] = (msgnum % 10) + '0';
			inheader = 1;
			found_status = 0;
			found_reply = 0;
		}
		if( inheader )mimehead(message);
		kanjicnv(message);
		if( inheader ){
			if( message[0] == ' ' || message[0] == 0x09 ){
				if( subcont )
					subcont = headcopy(headbuf + 79 - subcont, message, subcont);
				if( fromcont )
					fromcont = headcopy(headbuf + 39 - fromcont, message, fromcont);
			}
			else if( memcmp(message, "Subject: ", 9) == 0 ){
				subcont = headcopy(headbuf + 39, message + 9, 40);
				fromcont = 0;
			}
			else if( memcmp(message, "Reply-To: ", 10) == 0 ){
				memset(headbuf + 3, ' ', 36);
				fromcont = headcopy(headbuf + 3, message + 10, 36);
				subcont = 0;
				found_reply = 1;
			}
			else if( memcmp(message, "From: ", 6) == 0 ){
				if( found_reply == 0 ){
					fromcont = headcopy(headbuf + 3, message + 6, 36);
					subcont = 0;
				}
			}
			else if( memcmp(message, "Status: ", 8) == 0 ){
				strcpy(message, "Status: RO\n");
				found_status = 1;
				subcont = 0;
				fromcont = 0;
			}
			else{
				subcont = 0;
				fromcont = 0;
			}
		}
		else{
			if( memcmp(message, "From ", 5) == 0 )putc('>', fpmbox);
		}
		fputs(message, fpmbox);
		filetop = 0;
	}
	/* fputs("\n", fpmbox) */
	fputs(headbuf, stdout);
	fclose(fpmbox);
	fclose(fpwbox);

	/*
	 * delete messages from the server
	 */
#ifdef	DOSFILESYSTEM
	usr_reply = UserReply("Delete from Server? (Y/N)", 1);
#else
	fprintf(stderr, "Delete from Server? (Y/N)");
	gets(message);
	usr_reply = message;
#endif
	if( usr_reply[0] == 'Y' || usr_reply[0] == 'y' ){
		for( msgnum = 0 ; msgnum < msgamount ; ++msgnum ){
			sprintf(message, "DELE %d", msgnum + 1) ;
			SendCommand();
			if( GetReply() ){
				fputs("Cannot delete : ", stderr);
				fputs(message, stderr);
				putc('\n', stderr);
				ClosePopServer(1);
				exit(126);
			}
		}
#ifdef	KEEPMANY
		strcpy(message, "STAT");
		SendCommand();
		if( GetReply() ){
			fputs("cannot get status : ", stderr);
			fputs(message, stderr);
			putc('\n', stderr);
			ClosePopServer(1);
			exit(126);
		}
		msgamount = atoi(strtok(message, " "));
		if( msgamount != 0 ){
			fprintf(stderr, "Remaining Message Amount: %d \n", msgamount);
		}
#endif
	}
	ClosePopServer(1);
	return	0;
}

/*
** mimehead
**
** Decode In-header MIME-B/Q (Base64 / Quoted Printable) line by line
*/
static	void
mimehead(char *line){
int	cnvbuf[4];
char	*ptr1 = line;
char	*ptr2 = line;
char	*startptr;
int	i, codeway;

	while( *ptr1 ){
		/* Skip until MIME-B/Q Start Sequence */
		if( *ptr1 != '=' ){
			*ptr2++ = *ptr1++;
			continue;
		}
		ptr1++;
		while( *ptr1 == '=' ){
			*ptr2++ = '=';
			ptr1++;
		}
		if( *ptr1 == 0 ){
			*ptr2++ = '=';
			break;
		}
		if( *ptr1 != '?' ){
			*ptr2++ = '=';
			*ptr2++ = *ptr1++;
			continue;
		}
		startptr = ++ptr1;
		while( *ptr1 != '?' ){
			if( *ptr1 == 0 )break;
			if( *ptr1 <= ' ' )break;
			if( *ptr1 >= 0x7f )break;
			ptr1++;
		}
		if( *ptr1 != '?' ){
			*ptr2++ = '=';
			*ptr2++ = '?';
			while( startptr <= ptr1 ){
				*ptr2++ = *startptr++;
			}
			ptr1++;
			continue;
		}
		ptr1++;
		if( *ptr1 != 'B' && *ptr1 != 'b' && *ptr1 != 'Q' && *ptr1 != 'q' ){
			*ptr2++ = '=';
			*ptr2++ = '?';
			while( startptr <= ptr1 ){
				*ptr2++ = *startptr++;
			}
			ptr1++;
			continue;
		}
		codeway = *ptr1;
		ptr1++;
		if( *ptr1 != '?' ){
			*ptr2++ = '=';
			*ptr2++ = '?';
			while( startptr <= ptr1 ){
				*ptr2++ = *startptr++;
			}
			ptr1++;
			continue;
		}
		ptr1++;

		/* Decode MIME-B */
		if( codeway == 'B' || codeway == 'b' ){
			i = 0;
			codeway = 'B';
			while( codeway == 'B' ){
				if( *ptr1 >= 'A' && *ptr1 <= 'Z' ){
					cnvbuf[i++] = *ptr1 - 'A';
					ptr1++;
				}
				else if( *ptr1 >= 'a' && *ptr1 <= 'z' ){
					cnvbuf[i++] = *ptr1 + 26 - 'a';
					ptr1++;
				}
				else if( *ptr1 >= '0' && *ptr1 <= '9' ){
					cnvbuf[i++] = *ptr1 + 52 - '0';
					ptr1++;
				}
				else if( *ptr1 == '+' ){
					cnvbuf[i++] = 62;
					ptr1++;
				}
				else if( *ptr1 == '/' ){
					cnvbuf[i++] = 63;
					ptr1++;
				}
				else{
					codeway = *ptr1;
				}
				if( i >= 4 || codeway != 'B' ){
					if( i >= 2 )*ptr2++ = ( cnvbuf[0] << 2 ) | ( cnvbuf[1] >> 4 );
					if( i >= 3 )*ptr2++ = ( cnvbuf[1] << 4 ) | ( cnvbuf[2] >> 2 );
					if( i >= 4 )*ptr2++ = ( cnvbuf[2] << 6 ) | cnvbuf[3];
					i = 0;
				}
			}

			/* Skip MIME-B Terminate Sequence */
			while( *ptr1 == '=' )ptr1++;
			if( *ptr1 == '?' )ptr1++;
			if( *ptr1 == '=' )ptr1++;
		}

		/* Decode MIME-Q */
		else while( *ptr1 ){
			if( ptr1[0] == '?' && ptr1[1] == '=' ){
				ptr1 += 2;
				break;
			}
			if( *ptr1 == '=' ){
				if( ptr1[1] >= '0' && ptr1[1] <= '9' ){
					*ptr2 = ptr1[1] - '0';
				}
				else if( ptr1[1] >= 'A' && ptr1[1] <= 'F' ){
					*ptr2 = ptr1[1] - 'A' + 10;
				}
				else if( ptr1[1] >= 'a' && ptr1[1] <= 'f' ){
					*ptr2 = ptr1[1] - 'a' + 10;
				}
				else{
					*ptr2++ = *ptr1++;
					continue;
				}
				if( ptr1[2] >= '0' && ptr1[2] <= '9' ){
					*ptr2 *= 16;
					*ptr2 += ptr1[2] - '0';
				}
				else if( ptr1[2] >= 'A' && ptr1[2] <= 'F' ){
					*ptr2 *= 16;
					*ptr2 += ptr1[2] - 'A' + 10;
				}
				else if( ptr1[2] >= 'a' && ptr1[2] <= 'f' ){
					*ptr2 *= 16;
					*ptr2 += ptr1[2] - 'a' + 10;
				}
				else{
					*ptr2++ = *ptr1++;
					continue;
				}
				ptr1 += 3;
				ptr2++;
			}
			else{
				*ptr2++ = *ptr1++;
			}
		}
	}
	*ptr2 = 0;
}

/*
** kanjicnv
**
** Convert ISO-2022-JP into Shift_JIS or EUC_JP
** (accepts some non-JP ISO-2022 Codes)
**
** Any GR Code (including Shift_JIS and various EUC codes) 
** will be transparent. 
** New Line will cancel all Shifts and Directions. 
** Single-Byte Shift-Out will transfer all Non-X0208 Codes 
** to GR (in Shift_JIS mode) or G2 (in EUC_JP mode). 
** Any Double-Byte Shift Codes will be neglected (transparent). 
** Supplementary Kanji and Non-Japanese Multi-byte Codes will be 
**  transparent except when Single-Byte Shift-Out takes effect. 
** Direction of any Single-Byte Codes except JIS-X0201 Kana (including 
**  'Mis-Printed Version' i.e. being graphic set no. 7 instead of no. 9) 
**  will be assumed to be 'ASCII' 
** Any direction to G1, G2, and G3 will be neglected (transparent). 
*/
static	void
kanjicnv(char *line){
unsigned char	*ptr1 = (unsigned char *)line;
unsigned char	*ptr2 = (unsigned char *)line;
int	g0kana = 0;
int	g0kana0;
int	g1kana = 0;
int	kanji = 0;
unsigned	i, len;

#ifdef	EUC_JP
	ptr1 += strlen(ptr2) + 1;
	strcpy(ptr1, ptr2);
#endif

	while( *ptr1 ){
		if( *ptr1 == 0x1b /* ESC */ ){
			if( ptr1[1] == '[' ){
				for( len = 2 ; (ptr1[len] & 0xe0) == 0x20 ; len++ );
			}
			else{
				for( len = 1 ; (ptr1[len] & 0xf0) == 0x20 ; len++ );
			}

			if( ptr1[1] == '$' ){
				if( (len == 2 || ptr1[2] == '(')
				 && (ptr1[len] == '@' || ptr1[len] == 'B') ){
					g0kana = 0;
					kanji = 1;
					ptr1 += len + 1;
				}
				else{
					for( i = 0 ; i < len ; i++ ){
						*ptr2++ = *ptr1++;
					}
					if( *ptr1 > ' ' && *ptr1 < 0x7f ){
						*ptr2++ = *ptr1++;
						g0kana = 0;
						kanji = 0;
					}
				}
			}
			else if( ptr1[1] == '(' ){
				if( ptr1[len] <= ' ' || ptr1[len] >= 0x7f ){
					for( i = 0 ; i < len ; i++ ){
						*ptr2++ = *ptr1++;
					}
				}
				else{
					g0kana0 = g0kana;
					g0kana = (ptr1[len] == 'I' || ptr1[len] == 'G');
					if( kanji || (g0kana != g0kana0) ){
						ptr1 += len + 1;
					}
					else for( i = 0 ; i <= len ; i++ ){
						*ptr2++ = *ptr1++;
					}
					kanji = 0;
				}
			}
			else{
				for( i = 0 ; i < len ; i++ ){
					*ptr2++ = *ptr1++;
				}
				if( *ptr1 > ' ' && *ptr1 < 0x7f )*ptr2++ = *ptr1++;
			}
		}
		else if( *ptr1 == 0x0f /* SI */ ){
			g1kana = 0;
			ptr1++;
		}
		else if( *ptr1 == 0x0e /* SO */ ){
			g1kana = 1;
			ptr1++;
		}
		else if( *ptr1 <= ' ' || *ptr1 >= 0x7f ){
			*ptr2++ = *ptr1++;
		}
		else if( g0kana || g1kana ){
#ifdef	EUC_JP
			*ptr2++ = 0x8e; /* SS2 */
			*ptr2++ = *ptr1++;
#else
			*ptr2++ = *ptr1++ ^ 0x80;
#endif
		}
		else if( kanji ){
			if( ptr1[1] <= ' ' || ptr1[1] >= 0x7f ){
				*ptr2++ = *ptr1++;
			}
			else{
#ifdef	EUC_JP
				*ptr2++ = *ptr1++ ^ 0x80;
				*ptr2++ = *ptr1++ ^ 0x80;
#else
				ptr1[0] -= 0x21;
				if( ptr1[0] & 0x01 )ptr1[1] += ( 0x7f - 0x21 );
				ptr1[1] += ( 0x40 - 0x21 );
				ptr1[0] >>= 1;
				ptr1[0] += 0x81;
				if( ptr1[0] > 0x9f )ptr1[0] |= 0x40;
				if( ptr1[1] > 0x7e )ptr1[1]++;
				*ptr2++ = *ptr1++;
				*ptr2++ = *ptr1++;
#endif
			}
		}
		else{
			*ptr2++ = *ptr1++;
		}
	}
	*ptr2 = 0;
}

/*
** headcopy
**
** Copy the head of specified header field 
**  to the fixed-length field for display.
*/
static	int
headcopy(char *dest, char *src, int len){

	while( (unsigned char)*src <= ' ' ){
		src++;
		if( *src == 0 )return	len;
	}
	*dest++ = ' ';
	len--;
	while( len ){
		if( *src == 0 )break;
		if( ((*src & 0xe0) == 0x80)
		 || ((*src & 0xe0) == 0xe0) ){
			if( len == 1 ){
				*dest = ' ';
				len = 0;
			}
			else{
				*dest++ = *src++;
				*dest++ = *src++;
				len -= 2;
			}
		}
		else if( (unsigned char)*src >= ' ' ){
			*dest++ = *src++;
			len--;
		}
		else{
			src++;
		}
	}
	return	len;
}

/*
** ConnectPopServer
**
** The procedure to connect local and remote with socket interface.
*/
static	int
ConnectPopServer(char *hostname){
struct sockaddr_in server;
struct servent	*sp;
struct hostent	*hp;
int	sock;

	memset((char *)&server, 0, sizeof(server));
	server.sin_addr.s_addr = inet_addr(hostname);
	if( server.sin_addr.s_addr != ~0UL ){
		server.sin_family = AF_INET;
	}
	else{
		hp = gethostbyname(hostname);
		if( hp == NULL ){
			fprintf(stderr, "unknown host: %s\n", hostname);
			return	-1;
		}
		server.sin_family = hp->h_addrtype;
		memcpy((char *)&server.sin_addr, hp->h_addr, (size_t)hp->h_length);
	}
	sp = getservbyname(PROTOCOL, "tcp");
	if( sp == NULL ){
		fprintf(stderr, "unknown protocol: %s\n", PROTOCOL);
		return	-1;
	}
	server.sin_port = sp->s_port;
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if( sock < 0 ){
		perror("socket");
		return	-1;
	}
	if( connect( sock, (struct sockaddr *)&server, sizeof(server)) < 0 ){
		perror("connect");
		return	-1;
	}
#ifdef	MSDOS
	if( fcntl(sock, F_SETFL, O_NDELAY) ){
		perror("fcntl");
		return	-1;
	}
#endif
	return	sock;
}

/*
** ClosePopServer
**
** close several resources to exit peacefully.
*/
static	void
ClosePopServer(int waitreply){

	strcpy(message, "QUIT");
	SendCommand();
	if( waitreply ){
		if( GetReply() ){
			fputs("cannot quit safely : ", stderr);
			fputs(message, stderr);
			putc('\n', stderr);
			exit(126);
		}
	}
#ifdef	WIN32
	closesocket(sock);
#else
	close(sock);
#endif
}

/*
** downloadEmail
**
** The procedure of download a Email from remote. 
*/
static	void
downloadEmail(int msgnum){
static	char	*mboxhead;

	sprintf(message, "RETR %d", msgnum);
	SendCommand();
	if( GetReply() ){
		fputs("Cannot download : ", stderr);
		fputs(message, stderr);
		putc('\n', stderr);
		ClosePopServer(1);
		exit(126);
	}
	if( verbose )fprintf(stderr, "trying to get message %d...", msgnum);

	/* Check if the First Line is the `mbox' Header */
	if( ReceiveCommand() == NULL ){
		fprintf(stderr, "FAILURE in download message %d\n", msgnum);
		ClosePopServer(1);
		exit(126);
	}
	if( (memcmp(message, "From ", 5) != 0) &&
	    (memcmp(message, ">From ", 6) != 0) ){
		/* Make up `mbox' Header */
		mboxhead = message + strlen(message) + 1;
		fputs("From ", fpwbox);
		fputs(hostname, fpwbox);
		timestamp(mboxhead);
		fputs(mboxhead, fpwbox);
		putc('\n', fpwbox);
	}
	if( isEOM() ){
		putc('\n', fpwbox);
		if( verbose ){
			fprintf(stderr, "Null Message Body\n");
		}
		return;
	}
	fputs(message
	      + (((memcmp(message, "..", 2) == 0) ||
	          (memcmp(message, ">From ", 6) == 0)) ? 1 : 0), fpwbox);
	putc('\n', fpwbox);

	for(;;){
		if( ReceiveCommand() == NULL ){
			fprintf(stderr, "FAILURE in download message %d\n", msgnum);
			ClosePopServer(1);
			exit(126);
		}
		if( isEOM() )
			break;
		if( memcmp(message, "From ", 5) == 0 )putc('>', fpwbox);
		fputs(message + (memcmp(message, "..", 2) ? 0 : 1), fpwbox);
		putc('\n', fpwbox);
	}
	putc('\n', fpwbox);
	if( verbose ){
		fprintf(stderr, "SUCCEEDED\n");
	}
	return;
}

/*
** isEOM
**
** checks end of message. According to RFC-POP3, EOM mark is '.'.
*/
static	int
isEOM(void){
	return	( message[0] == '.' && message[1] == '\0' );
}

/*
** GetReply
**
** check ACK taken via ReceiveCommand() and return pointer
** of the content of message. The message may includes '+OK'
** or '-ERR' to indicate the status.
*/
static	int
GetReply(void){
	if( ReceiveCommand() == NULL ){
		return	-1;
	}
	if( verbose ){
		fputs(message, stderr);
		putc('\n', stderr);
	}
	if( strncmp(message, "+OK ", 4) == 0 ){
		memcpy(message, message+4, strlen(message)+1);
		return	0;
	}
	else if( strncmp(message, "+OK", 4) == 0 ){
		message[0] = 0;
		return	0;
	}
	else if( strncmp(message, "-ERR ", 5) == 0 ){
		memcpy(message, message+5, strlen(message)+1);
		return	1;
	}
	else if( strncmp(message, "-ERR", 5) == 0 ){
		message[0] = 0;
		return	1;
	}
	else{
		return	2;
	}
}

/*
** SendCommand
**
** low level network function to sends RFC1225 commands to my pop
** server.
*/
static	void
SendCommand(void){
	strcat(message, "\r\n");
	send(sock, message, strlen(message), 0);
	if( verbose ){
		message[strlen(message) - 2] = '\n';
		message[strlen(message) - 1] = 0;
		fputs(message, stderr);
	}
}

/*
** ReceiveCommand
**
** low level network function to receives result messages from my
** pop server. 
*/
static	char
*ReceiveCommand(void){
static	int	recvret = 0, reststart = 0;
int	linelength = 0;
	for(;;){
		while( recvret > reststart ){
			message[linelength] = sockbuf[reststart];
			reststart++;
			if( (message[linelength] == '\n') &&
			    (linelength > 0) &&
			    (message[linelength - 1] == '\r') ){
				message[linelength - 1] = 0;
				return	message;
			}
			if( message[linelength] == 0 ){
				return	message;
			}
			if( ++linelength >= LINEBUFSIZE - 1 ){
				message[linelength] = 0;
				return	message;
			}
		}
		reststart = 0;
		if( (recvret = recv(sock, sockbuf, SOCKBUFSIZE, 0)) == 0 ){
			fprintf(stderr, "recv: Unexpected Disconnect\n");
			return	NULL;
		}
		else if( recvret < 0 ){
#ifdef	WIN32
			if( WSAGetLastError() != EWOULDBLOCK ){
#else
			if( errno != EWOULDBLOCK ){
#endif
				perror("recv");
				return	NULL;
			}
#ifdef	MSDOS
			else if( (bdos(6, 0xff, 0) & 0xff) == 0x03 ){
				fprintf(stderr, "^C\n");
				return	NULL;
			}
#endif
		}
	}
}

/*
** mboxpurge
**
** delete mbox files before exit
*/
static	void
mboxpurge(void){
	fclose(fpmbox);
	fclose(fpwbox);
	unlink(wboxname);
	if( *mboxopen != 'a' )unlink(mboxname);
}

/*
** timestamp
**
** Make a Time Stamp in the `mbox' Header
**
** This procedure can be written as 
**   #define TIMEFORM " %a %b %e %T %Z %Y"
**   time_t timebuf;
**   time(&timebuf);
**   strftime(mboxhead, LINEBUFSIZE - (mboxhead - message), TIMEFORM, localtime(&timebuf));
** but some of the author's targets do not support `strftime'. 
*/
static	void
timestamp(char *target){
static	char	*wday[7] 
	= {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"};
static	char	*month[12] 
	= {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
time_t	times;
struct	tm	*timebuf;

	time(&times);
	timebuf = localtime(&times);

	target[0] = ' ';
	memcpy(target + 1, wday[timebuf -> tm_wday], 3);
	target[4] = ' ';
	memcpy(target + 5, month[timebuf -> tm_mon], 3);
	target[8] = ' ';
	setdecimal(target + 9, timebuf -> tm_mday, 2, 0);
	target[11] = ' ';
	setdecimal(target + 12, timebuf -> tm_hour, 2, 1);
	target[14] = ':';
	setdecimal(target + 15, timebuf -> tm_min, 2, 1);
	target[17] = ':';
	setdecimal(target + 18, timebuf -> tm_sec, 2, 1);
	target[20] = ' ';
	target[21] = 0;
	strcat(target, tzname[0]);
	strcat(target, " ");
	setdecimal(target + strlen(target), timebuf -> tm_year + 1900 , 4, 0);

	return;
}

/*
** setdecimal
**
** Change Integer to Decimal Numerals
**
** subroutine of `timestamp'
*/
static	void
setdecimal(char *target, int value, int column, int fillzero){
	target[column] = 0;
	while( column ){
		column--;
		if( value || fillzero ){
			target[column] = '0' + (value % 10);
			value /= 10;
		}
		else{
			target[column] = ' ';
		}
	}
}

#ifdef	DOSFILESYSTEM
/*
** UserReply
**
** Reply by User
**
*/
static	char
*UserReply(const char *prompt, int visible){
static	char	result[TYPEBUFSIZE];
int	c;
int	ptr = 0;

	while( *prompt ){
#ifdef	MSDOS
		bdos(6, *prompt, 0);
#else	/* WIN32 */
		_putch(*prompt);
#endif
		prompt++;
	}
#ifdef	MSDOS
	while( (c = bdos(6, 0xff, 0) & 0xff) != 0x0d ){
#else	/* WIN32 */
	while( (c = _getch()) != 0x0d ){
#endif
		if( c >= ' ' && c < 0xff && ptr < TYPEBUFSIZE ){
			result[ptr] = c;
			if( visible ){
#ifdef	MSDOS
				bdos(6, c, 0);
#else	/* WIN32 */
				_putch(c);
#endif
			}
			ptr++;
		}
		else if( c == 0x08 || c == 0x7f && ptr ){
			if( visible ){
#ifdef	MSDOS
				bdos(6, 0x08, 0);
				bdos(6, 0x20, 0);
				bdos(6, 0x08, 0);
#else	/* WIN32 */
				_putch(0x08);
				_putch(0x20);
				_putch(0x08);
#endif
			}
			ptr--;
		}
	}
	result[ptr] = 0;
#ifdef	MSDOS
	bdos(6, '\r', 0);
	bdos(6, '\n', 0);
#else	/* WIN32 */
	_putch('\r');
	_putch('\n');
#endif
	return	result;
}
#endif
